import os

DEFAULT_LEARNING_RATE = 0.001
DEFAULT_TF_MODELFILE = "model.tf"
DEFAULT_TF_CHECKPOINT_DIR = "model.tf"
DEFAULT_TF_CHECKPOINT = os.path.join(DEFAULT_TF_CHECKPOINT_DIR, "checkpoint")
# DEFAULT_TF_GENERATOR_MODELFILE = "model_generator.tf"
# DEFAULT_TF_DISCRIMINATOR_MODELFILE = "model_discriminator.tf"
